#==============================================================================
#
#  $Id: dopy.py,v 1.5 1999/04/24 21:51:30 mike Exp $
#
"""
   The DOPY main module.  This houses all of the central classes and base
   classes that comprise the core DOPY system.
"""
#
#  Copyright (C) 1999 Michael A. Muller
#
#  Permission is granted to use, modify and redistribute this code,
#  providing that the following conditions are met:
#
#  1) This copyright/licensing notice must remain intact.
#  2) If the code is modified and redistributed, the modifications must 
#  include documentation indicating that the code has been modified.
#  3) The author(s) of this code must be indemnified and held harmless
#  against any damage resulting from the use of this code.
#
#  This code comes with ABSOLUTELY NO WARRANTEE, not even the implied 
#  warrantee of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
#
#  $Log: dopy.py,v $
#  Revision 1.5  1999/04/24 21:51:30  mike
#  Bug fixes: context object was never getting released (causing the protocol to
#  never be released).  Switched to traceback module for displaying tracebacks
#  from mainloop CLI.
#
#  Revision 1.4  1999/04/22 00:30:19  mike
#  Added the Context facility to the hub.
#
#
#==============================================================================


"""
   Limitations
   -----------
   
   At this time, output parameters are not supported (except by passing in
   remote objects).
   
   XXX should protocol and protocol info objects be one and the same?
   Or would this only pollute the protocol interface?  There is a speed
   advantage to this - but I think that protocol info may need to manage
   the underlying connection object.
   
"""

import thread, string

class Error(Exception):
   """
      Base class for all DOPY errors.
   """
   pass

class InvalidMethod(Error):
   """
      Raised when an invalid method name is cited in a request.
   """
   pass

class InvalidObjectKey(Error):
   """
      Raised when an invalid object key is cited as the target for a 
      request.
   """
   pass

class BadResponseError(Error):
   """
      Raised when a response is received with an ID other than that expected.
   """
   pass

class ProtocolTerminatedError(Error):
   """
      Raised when a protocol is terminated.
   """
   pass

class _Abort(Exception):
   """
      Internal error raise when a failed precondition aborts request
      processing.
   """
   pass

class RemoteObject:
   
   """
      The proxy object.
      
      The following public variables exist:
      
      dopy_protoSpec::
         An instance of *Protocol* defining the protocol spec for the
         object.
      dopy_key::
         The key for the object.  Should be unique within the protocol
         spec.
   """
   
   def __init__(self, protoSpec, key):
      self.dopy_protoSpec = protoSpec
      self.dopy_key = key

   def __getstate__(self):
      dict = self.__dict__.copy()
      dict['dopy_protoSpec'] = self.dopy_protoSpec.getProtocolId()
      return dict

   def __setstate__(self, dict):
      self.__dict__ = dict
      self.dopy_protoSpec = _hub.getProtocol(self.dopy_protoSpec)
   
class RemoteMethod:
   """
      A remote method stores information about an operation that can
      be performed on a remote object.
   """
   
   def __init__(self, target, name):
      self.target = target
      self.name = name
   
   def __call__(self, *parms, **kwParms):
      req = Request(self.target, self.name, parms, kwParms)
      response = _hub.processRequest(req)
      if response.exception is not None:
         raise response.exception
      else:
         return response.returnValue

class DynamicRemoteObject(RemoteObject):
   
   """
      A dynamic remote object has no facility for local validation of
      method invocations: it has no proxy.  Therefore, it is the
      easiest way to execute things remotely.  Attribute errors are very
      costly, though.
   """
   
   def __getattr__(self, attr):
      return RemoteMethod(self, attr)

class Request:

   """
      Contains all of the information necessary to invoke a remote method.
      Public variables:
      /id/::
         The message id which is used to associate the request with its
         response (set by the hub before the request is sent).
      /target/::
         The object key that the method is being invoked upon.
      /method/::
         The name of the method.
      /parms/::
         The parameter sequence.
      /kwParms/::
         The keyword parameters.
   """
   
   def __init__(self, target, method, parms, kwParms):
      self.id = None
      self.target = target
      self.method = method
      self.parms = parms
      self.kwParms = kwParms

   def __getstate__(self):
      # replace the object with its key so that we don't get into
      # an infinitely recursive loop looking for __getinitargs__ with 
      # __getattr__
      dict = self.__dict__.copy()
      dict['target'] = self.target.dopy_key
      return dict

class Response:
   """   
      Response to a Request.  Public variables:
      /id/::
         Response id.  This is the same id as that of the matching request.
      /returnValue/::
         The return value of the remote method.
      /exception/::
         The exception raise by the remote method (*None* if no
         exception was raised).
   """
   
   def __init__(self, id, returnValue = None, exception = None):
      self.id = id
      self.returnValue = returnValue
      self.exception = exception

class _AwaitingResponse:
   
   """
      Instances of this class track pending requests.  They are managed by
      the Hub in the sendRequest() method.
      
      Public variables:
      /id/::
         The key that this object is stored under.  The same id as is used in 
         the matching Request and Response objects.
      /response/::
         Allows us to store the response so that it can be returned to the
         caller.
      /threaded/::
         A class level variable that is used to indicate whether we are using
         the threaded model or not.
   """
   
   threaded = 1
   
   def __init__(self, id):
      """
         Constructs a new instance using the given id.
      """
      if self.threaded:
         self.__wait = thread.allocate_lock()
         self.__wait.acquire()
      self.id = id
      self.response = None
   
   def responseReceived(self, response):
      """
         Used to indicate that a response has been received and to set the
         response that was received.  Any threads waiting on getResponse()
         will continue.
      """
      self.response = response
      if self.threaded:
         self.__wait.release()
   
   def getResponse(self):
      """
         Waits for a response to be received and then returns it.
      """
      if self.threaded:
         self.__wait.acquire()
      else:
         self.response = self.hub.getResponse(self.id)
      return self.response

class FileEventSource:
   
   """
      This is any kind of object that can be used within the select() method
      as a source of low-level events.
   """
   
   def fileno(self):
      """
         Derived classes should implement this to provide a method that
         returns a file number for the object, suitable for use in the
         select() function.
      """
      raise ImplementedBySubclassError()

   def isActive(self):
      """
         Returns true for as long as the server is active.
      """
      raise ImplementedBySubclassError()
   
   def shutdown(self):
      """
         Shuts down the connection.
      """
      raise ImplementedBySubclassError()

class ProtocolServer(FileEventSource):
   
   """
      Protocol servers handle incoming connections on a particular protocol.
   """
   
   def acceptConnection(self):
      """
         Returns a new connection for the server.
      """
      raise ImplementedBySubclassError()   
      

class Protocol(FileEventSource):
      
   """
      Handles the nitty-gritty details of sending and recieving requests
      and responses across a communication protocol.
      
      Since protocols must be unpickleable, there may be multiple instances
      of the same protocol object.  Therefore, there should generally be
      some object that the protocol can locate within its __setattr__ method
      that actually houses the single instance of the connection that
      the multiple protocol objects refer to.
   """
   
   def getNext(self):
      """
         Derived classes must implement this to get the next object
         received on the protocol.  Should return *None* if the
         connection closes.
      """
      raise ImplementedBySubclassError()

   def send(self, msg):
      """
         Sends a request or response message through the protocol.
      """
      raise ImplementedBySubclassError()

   def getProtocolId(self):
      """
         Derived classes must use implement this to return a unique protocol
         id.  Protocol ids should be strings beginning with a type
         identifier followed by a colon (e.g. "tcp:") and followed by a value
         that uniquely identifies the connection.
      """
      raise ImplementedBySubclassError()

class ProtocolInfo:

   """
      Protocol info manages the current message id and a list of requests.
      Public variables:
      /requests/::
         A dictionary mapping message ids to request locks.
      /proto/::
         The Protocol object itself.
   """
      
   def __init__(self, protocol):
      """
         Constructs a ProtocolInfo object from the Protocol itself.
      """
      self.__lk = thread.allocate_lock()
      self.__counter = 0
      self.requests = {}
      self.proto = protocol
   
   def getNextId(self):
      """
         Returns the next available message id.
      """
      self.__lk.acquire()
      rc = self.__counter
      self.__counter = self.__counter + 1
      self.__lk.release()
      return rc

class Context:
   """
      Instances of this class manage information about the caller context
      which may be used during the processing of a remote method.
   """

   def __init__(self, protocol):
      self.__protocol = protocol
   
   def protocol(self):
      """
         Returns the protocol which was used to communicate to the hub.
      """
      return self.__protocol

class Hub:
   
   """
      The Hub is the equivalent of the ORB - a central context for everything
      in the distributed object system.
      
      The Hub has a *registry* which can be as simple as a dictionary
      or as sophisticated as a CORBA NamingContext.  It must provide
      the __getitem__ and __setitem__ methods to dereference and assign
      keys to objects.
   """
   
   def __init__(self, registry):
      self.__registry = registry
      self.__server = None
      self.__threaded = 1
      self.__protocolInfo = {}
      if self.__threaded:
         self.__context = {}
      else:
         self.__context = None

   def isThreaded(self):
      return self.__threaded
   
   def addServer(self, server):
      self.__server = server
      if self.__threaded:
         thread.start_new_thread(self.__serviceConnections, (server,))

   def addObject(self, key, object):
      self.__registry[key] = object

   def addProtocol(self, proto):
      """
         This function should generally be called by the protocol constructor
         to register itself with the hub.
      """
      self.__protocolInfo[proto.getProtocolId()] = ProtocolInfo(proto)
      if self.__threaded:
         thread.start_new_thread(self.__serviceClient, (proto,))

   def getProtocol(self, protoId):
      """
         This is used to get or create a protocol from its id.  Returns the
         protocol if it already exists, creates it if it doesn't.
      """
      try:
         return self.__protocolInfo[protoId].proto
      except KeyError, ex:
         global protocolFactory
         return protocolFactory.create(protoId)

   def getContext(self):
      """
         Returns the current context object.  This is only available from
         within a remote method call.  If it is called from somewhere
         other than a remotely invoked method, *None* is returned.
      """
      if self.__threaded:
         return self.__context.get(thread.get_ident())
      else:
         return self.__context

   def processRequest(self, request):
      """
         Processes a request (an instance of @Request) and returns
         the response (a @Response).
      """
      # get the protocol and protocol info objects
      proto = request.target.dopy_protoSpec
      protoInfo = self.__protocolInfo[proto.getProtocolId()]
      
      # give the request a unique id
      request.id = protoInfo.getNextId()
      
      # build the object that we will use to get our response
      resp = _AwaitingResponse(request.id)
      protoInfo.requests[request.id] = resp
      
      # send the request
      proto.send(request)
      
      # get and return the response
      return resp.getResponse()
   
   def __serviceConnections(self, server):
      while server.isActive():
         server.acceptConnection()

   def __connectionClosed(self, protoId, protoInfo):
      """
         Called when a connection is closed.
      """
      for waiter in protoInfo.requests.values():
         resp = Response(waiter.id, 
                         exception = ProtocolTerminatedError(protoId)
                         )
         waiter.responseReceived(resp)
      del self.__protocolInfo[protoId]
   
   def __serviceClient(self, clientConnection):
      myid = clientConnection.getProtocolId()
      proto = self.__protocolInfo[myid]
      while clientConnection.isActive():      
         request = clientConnection.getNext()
         
         # if the getNext() method returned None, the connection has gone
         # away: clean up and exit
         if not request:
            self.__connectionClosed(myid, proto)
            break
            
         elif isinstance(request, Response):
            proto.requests[request.id].responseReceived(request)
         else:
            response = self.__processRequest(request, clientConnection)
            clientConnection.send(response)
   
   def __processRequest(self, request, protocol):
      try:
         try:
            obj = self.__registry[request.target]
         except KeyError, ex:
            response = Response(request.id, exception = InvalidObjectKey(ex))
            raise _Abort()
            
         try:
            meth = getattr(obj, request.method)
         except AttributeError, ex:
            response = Response(request.id, exception = InvalidMethod(ex))
            raise _Abort()
         
         self.__setContext(Context(protocol))
         try:
            retval = apply(meth, request.parms, request.kwParms)
            response = Response(request.id, returnValue = retval)
         except Exception, ex:
            response = Response(request.id, exception = ex)
         self.__clearContext()
      except _Abort, ex:
         pass
      return response      

   def __selectLoop(self, msgId):
      wr = rd = er = [self.__server]
      while 1:
         wrx, rdx, erx = select.select(wr, rd, er)
         while rdx:
            s = rdx.pop()
            if isinstance(s, ProtocolServer):
               wr.append(s.acceptConnection())
            else:
               req = s.getNext()
               if isinstance(req, Response):
                  if req.id != msgId:
                     raise BadResponseError("message id %s received, %s "
                                             "expected." % msgId)
                  return req
               else:
                  resp = self.__processRequest(req, s)
                  s.send(resp)            

   def __setContext(self, context):
      if self.__threaded:
         self.__context[thread.get_ident()] = context
      else:
         self.__context = context
   
   def __clearContext(self):
      if self.__threaded:
         del self.__context[thread.get_ident()]
      else:
         self.__context = None
   
   def mainloop(self):
      if self.__threaded:
         import traceback
         while 1:
            inp = raw_input()
            if inp[:2] == '@q':
               return
            try:
               exec inp
            except SystemExit, ex:
               raise
            except Exception, ex:
               traceback.print_exc()
      else:
         self.__selectLoop()

   def _getResponse(self, responseId):
      """
         This function reenters the message processing loop and waits for a
	 particular response.

         If any response other that response should be received, a 
         @BadResponse exception will be raised.
      """
      assert not self.__threaded
      return self.__selectLoop(responseId)

class ProtocolFactory:
   """
      The protocol factory is a singleton that is always in existence - it
      manages the creation functions for different protocols.
      
      Protocols should each provide a factory method that creates the protocol
      from a list of strings - these being the fields (as divided by colons)
      of the protocol id.
   """

   def __init__(self):
      self.protos = {}
   
   def register(self, protoTypeId, protoFact):
      self.protos[protoTypeId] = protoFact
   
   def create(self, protoId):
      name = string.split(protoId, ':')
      return self.protos[name[0]](name)

protocolFactory = ProtocolFactory()

_hub = None

def getHub():
   global _hub
   if not _hub:
      _hub = Hub({})
   return _hub
